use age::{Recipient, secrecy::ExposeSecret};
use minisign::{PError, PublicKey};
use std::{io::Cursor, path::PathBuf, str::FromStr};

/// A cryptographic `Identity` consisting of an encryption and signing key
pub struct Identity {
    pub age: age::x25519::Identity,
    pub sign: minisign::KeyPair,
}

/// Save two parts into one.
///
/// # Errors
/// `part1` should NOT be bigger than `2^32`
fn save_parts(part1: &[u8], part2: &[u8]) -> Vec<u8> {
    let mut vec = Vec::new();

    let offset = part1.len() as u32;
    vec.extend_from_slice(&offset.to_le_bytes());

    vec.extend_from_slice(part1);
    vec.extend_from_slice(part2);

    vec
}

/// Load back two parts from one
fn load_parts(data: &[u8]) -> (Vec<u8>, Vec<u8>) {
    let offset_bytes: [u8; 4] = data[..4].try_into().unwrap();
    let offset = u32::from_le_bytes(offset_bytes) as usize;

    let part1 = data[4..4 + offset].to_vec();
    let part2 = data[4 + offset..].to_vec();

    (part1, part2)
}

impl Default for Identity {
    fn default() -> Self {
        Self::new()
    }
}

impl Identity {
    /// Generate a new `Identity`
    #[must_use]
    pub fn new() -> Self {
        let age = age::x25519::Identity::generate();
        let sign = minisign::KeyPair::generate_encrypted_keypair(Some(String::new())).unwrap();
        Self { age, sign }
    }

    /// Export public keys.
    #[must_use]
    pub fn public(&self) -> Persona {
        (self.pub_key_age(), self.pub_key_sign())
    }

    /// Get the age encryption public key
    #[must_use]
    pub fn pub_key_age(&self) -> String {
        self.age.to_public().to_string()
    }

    /// Get the minisign public key
    #[must_use]
    pub fn pub_key_sign(&self) -> String {
        self.sign.pk.to_box().unwrap().to_string()
    }

    /// Save the `Identity` to `dir`.
    ///
    /// `dir` will contain two files `age.key` and `sign.key`
    pub fn save<P: Into<PathBuf>>(&self, dir: P) {
        let dir = dir.into();
        let age_key = self.age.to_string();
        std::fs::write(dir.join("age.key"), age_key.expose_secret()).unwrap();

        let kbx = self.sign.sk.to_box(None).unwrap();

        std::fs::write(dir.join("sign.key"), kbx.into_string()).unwrap();
        log::info!("Saved identity to {}", dir.display());
    }

    /// Load an `Identity` from `dir`.
    ///
    /// `dir` should contain the files `age.key` and `sign.key`
    pub fn try_load<P: Into<PathBuf>>(dir: P) -> Option<Self> {
        let dir = dir.into();
        let age =
            age::x25519::Identity::from_str(&std::fs::read_to_string(dir.join("age.key")).ok()?)
                .unwrap();
        let kbx = minisign::SecretKeyBox::from_string(
            &std::fs::read_to_string(dir.join("sign.key")).ok()?,
        )
        .unwrap();
        let sign = kbx.into_secret_key(Some(String::new())).unwrap();
        let pk = minisign::PublicKey::from_secret_key(&sign).ok()?;
        let kp = minisign::KeyPair { pk, sk: sign };
        log::info!("Loaded identity from {}", dir.display());
        Some(Self { age, sign: kp })
    }

    /// Encrypt `data` for a `recipient`
    pub fn encrypt(&self, data: &[u8], recipient: &impl Recipient) -> Vec<u8> {
        // STEP 1 : Sign
        let signed = minisign::sign(
            None,
            &self.sign.sk,
            data,
            Some(&chrono::Utc::now().timestamp().to_string()),
            None,
        )
        .unwrap();
        let signed = save_parts(&signed.to_bytes(), data);

        // STEP 2 : Encrypt
        let enc = age::encrypt(recipient, &signed).unwrap();

        // STEP 3: Sign
        let signed = minisign::sign(None, &self.sign.sk, Cursor::new(&enc), None, None).unwrap();
        save_parts(&signed.to_bytes(), &enc)
    }

    /// Decrypt `data` and verify the signatures using `pk`
    pub fn decrypt(&self, data: &[u8], pk: &PublicKey) -> Result<Message, DecryptError> {
        // STEP 1 : Verify outer sign
        let (sig, data) = load_parts(data);
        let sign_box = minisign::SignatureBox::from_string(
            &String::from_utf8(sig).map_err(|_| DecryptError::Encoding)?,
        )
        .map_err(|_| DecryptError::OuterSign)?;
        if minisign::verify(pk, &sign_box, Cursor::new(&data), true, false, false).is_err() {
            return Err(DecryptError::OuterSign);
        }

        // STEP 2 : Decrypt
        let dec = age::decrypt(&self.age, &data).map_err(DecryptError::Decrypt)?;

        // STEP 3 : Verify inner sign
        let (sig, data) = load_parts(&dec);
        let sign_box = minisign::SignatureBox::from_string(
            &String::from_utf8(sig).map_err(|_| DecryptError::Encoding)?,
        )
        .map_err(|_| DecryptError::InnerSign)?;
        if minisign::verify(pk, &sign_box, Cursor::new(&data), true, false, false).is_err() {
            return Err(DecryptError::InnerSign);
        }

        let tc = sign_box
            .trusted_comment()
            .map_err(|_| DecryptError::Encoding)?;
        let timestamp: i64 = tc.parse().map_err(|_| DecryptError::Encoding)?;

        Ok(Message {
            payload: data,
            timestamp,
        })
    }
}

/// A signed encrypted message
pub struct Message {
    /// The unencrypted payload
    pub payload: Vec<u8>,
    /// A timestamp from the signature
    pub timestamp: i64,
}

/// General decryption error
#[derive(Debug)]
pub enum DecryptError {
    /// Decryption failed on the outer signature
    OuterSign,
    /// Decryption failed on the encrypted payload
    Decrypt(age::DecryptError),
    /// Decryption failed on the inner signature
    InnerSign,
    /// Wrong encoding
    Encoding,
}

/// A public persona `(AgePubKey, MinisignPubKey)`
type Persona = (String, String);

pub trait PersonaIdentity {
    fn sign_key(&self) -> Result<minisign::PublicKey, minisign::PError>;

    fn enc_key(&self) -> Option<age::x25519::Recipient>;
}

impl PersonaIdentity for Persona {
    /// Get the public sign key
    fn sign_key(&self) -> Result<minisign::PublicKey, PError> {
        let kbx = minisign::PublicKeyBox::from_string(&self.1)?;
        kbx.into_public_key()
    }

    /// Get the public encryption key
    fn enc_key(&self) -> Option<age::x25519::Recipient> {
        age::x25519::Recipient::from_str(&self.0).ok()
    }
}
